#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

__title__ = "p5"
__description__ = "Creative coding in Python"
__url__ = "https://p5py.github.io"
__version__ = "0.8.0"
__author__ = "Abhik Pal"
__author_email__ = "theabhikpal@gmail.com"
__license__ = " GNU GPLv3"
__copyright__ = "Copyright (C) 2017-2020 Abhik Pal"
